from typing import TextIO, Union

import yaml

try:
    from yaml import CSafeDumper as _SafeDumper
    from yaml import CSafeLoader as _SafeLoader
except ImportError:
    from yaml import SafeDumper as _SafeDumper
    from yaml import SafeLoader as _SafeLoader


def dump(data, stream=None):
    return yaml.dump(
        data,
        stream=stream,
        Dumper=_SafeDumper,
        default_flow_style=False,
        allow_unicode=True,
    )


def safe_load(filepath: Union[str, TextIO]):
    return yaml.load(filepath, Loader=_SafeLoader)



import os
import shlex
from subprocess import PIPE

from psutil import Popen


def run_command(cmd, data, location, chw, env=None):
    cmd_env = None
    if env:
        cmd_env = os.environ.copy()
        cmd_env.update(env)
    cwd = os.getcwd()
    if location is not None and chw is True:
        cwd = location
    elif location is not None and chw is False:
        cmd = "{0} {1}".format(cmd, location)
    r = Popen(
        shlex.split(cmd), stdout=PIPE, stdin=PIPE, stderr=PIPE, cwd=cwd, env=cmd_env
    )
    if data is None:
        output = r.communicate()[0].decode("utf-8")
    else:
        output = r.communicate(input=data)[0]
    return output



import math
from typing import Any, Dict

try:
    import numpy as np
except ImportError:
    np = None


def is_np_numerical_value_unsupported(value):
    return value is None or math.isnan(value) or math.isinf(value)


def sanitize_np_types(value: Any) -> Any:
    if isinstance(value, str):
        return value
    if is_np_numerical_value_unsupported(value):
        return None
    if isinstance(value, (int, float, complex, type(None))):
        return value
    if np and np.isnan(value):
        return None
    if np and isinstance(value, np.integer):
        return int(value)
    if np and isinstance(value, np.floating):
        return float(value)
    return value


def to_np(value: Any) -> Any:
    if hasattr(value, "detach"):  # Torch detach handling
        value = value.detach()
    if hasattr(value, "numpy"):  # Torch numpy handling
        value = value.numpy()
    if isinstance(value, np.ndarray):
        return value
    if np.isscalar(value):
        return np.array([value])


def calculate_scale_factor(tensor: Any):
    converted = tensor.numpy() if not isinstance(tensor, np.ndarray) else tensor
    return 1 if converted.dtype == np.uint8 else 255


def sanitize_dict(s: Dict) -> Dict:
    return {k: sanitize_np_types(v) for k, v in s.items()}



from enum import Enum
from typing import Iterable, List, Set, Tuple, Type, Union


def get_enum_value(e: Union[str, Enum]) -> str:
    if isinstance(e, Enum):
        return e.value
    return e


def enum_to_choices(enumeration: Type[Enum]) -> Iterable[Tuple[str, str]]:
    return tuple((e.value, e.value) for e in enumeration)


def enum_to_list(enumeration: Type[Enum]) -> List[str]:
    return [e.value for e in enumeration]


def enum_to_set(enumeration: Type[Enum]) -> Set[str]:
    return set(e.value for e in enumeration)


def values_to_choices(enumeration: Union[List, Set]) -> Iterable[Tuple[str, str]]:
    return tuple((e, e) for e in sorted(enumeration))


class EnumMixin:
    @classmethod
    def to_choices(cls) -> Iterable[Tuple[str, str]]:
        return enum_to_choices(cls)

    @classmethod
    def to_set(cls) -> Set[str]:
        return enum_to_set(cls)

    @classmethod
    def to_list(cls) -> List[str]:
        return enum_to_list(cls)

    @classmethod
    def members(cls) -> List[str]:
        return cls._member_names_


class PEnum(EnumMixin, Enum):
    pass



import logging
from collections.abc import Mapping
from typing import Any, List

try:
    import numpy as np
except ImportError:
    np = None


def to_list(
    value: Any,
    check_none: bool = False,
    check_dict: bool = False,
    to_unique: bool = False,
) -> List:
    def _to_unique(v):
        try:
            return list(dict.fromkeys(v))
        except Exception as e:
            logging.debug("Could not return unique value for list. Error %s", e)
            return list(v)

    if check_none and value is None:
        return []
    if isinstance(value, (list, tuple, set)):
        return _to_unique(value) if to_unique else list(value)
    if np and isinstance(value, np.ndarray):
        value = value.tolist()
        return _to_unique(value) if to_unique else value
    if check_dict and isinstance(value, Mapping):
        return list(value.items())
    return [value]



import json
from collections.abc import Mapping
from typing import Dict


def sanitize_value(d: Dict, handle_dict: bool = False):
    if isinstance(d, str):
        return d
    if not isinstance(d, Mapping):
        return json.dumps(d)
    if not handle_dict:
        return json.dumps(d)
    return {d_k: sanitize_value(d_v, handle_dict=True) for d_k, d_v in d.items()}


def sanitize_string_dict(d: Dict[str, str] = None):
    if isinstance(d, Mapping):
        return {d_k: sanitize_value(d_v, handle_dict=False) for d_k, d_v in d.items()}
    return d



import sys
from contextlib import contextmanager
from typing import Callable, List, Tuple

NEWLINES = ("\n", "\r", "\r\n")
INDENT_CHAR = " "
INDENT_STRINGS = []
STDOUT = sys.stdout.write
STDERR = sys.stderr.write


def _indent(depth: int = 0, quote: str = ""):
    """Indent util function, compute new indent_string"""
    if depth > 0:
        indent_string = "".join((str(quote), (INDENT_CHAR * (depth - len(quote)))))
    else:
        indent_string = "".join((("\x08" * (-1 * (depth - len(quote)))), str(quote)))

    if indent_string:
        INDENT_STRINGS.append(indent_string)


@contextmanager
def _indent_context():
    """Indentation context manager."""
    try:
        yield
    finally:
        # Remove indent
        INDENT_STRINGS.pop()


def indent(depth: int = 4, quote: str = ""):
    """Indentation manager, return an indentation context manager."""
    _indent(depth, quote)
    return _indent_context()


def split_many(string: str, delimiters: Tuple[str, ...]) -> List[str]:
    """Behaves str.split but supports tuples of delimiters."""
    delimiters = tuple(delimiters)
    if len(delimiters) < 1:
        return [string]
    final_delimiter = delimiters[0]
    for i in delimiters[1:]:
        string = string.replace(i, final_delimiter)
    return string.split(final_delimiter)


def puts(s: str = "", newline: bool = True, stream: Callable = STDOUT):
    """Prints given string to stdout."""
    if newline:
        s = split_many(s, NEWLINES)
        s = map(str, s)
        indentation = "".join(INDENT_STRINGS)

        s = (str("\n" + indentation)).join(s)

    outputs = "".join(("".join(INDENT_STRINGS), str(s), "\n" if newline else ""))
    stream(outputs)



import json
import sys
from typing import Optional

import click
import yaml
from clipped.humanize import humanize_attrs
from clipped.list_utils import to_list
from clipped.units_processors import to_percentage, to_unit_memory
from rich import box
from rich.console import Console
from rich.live import Live
from rich.markdown import Markdown
from rich.progress import (
    BarColumn,
    DownloadColumn,
    Progress,
    TaskProgressColumn,
    TextColumn,
    TimeElapsedColumn,
    TimeRemainingColumn,
    TransferSpeedColumn,
)
from rich.syntax import Syntax
from rich.table import Column, Table
from rich.theme import Theme


def dict_tabulate(dict_value, is_list_dict=False):
    if is_list_dict:
        headers = dict_value[0].keys() if dict_value else []
        table = Printer.get_table(*headers)
        for d in dict_value:
            table.add_row(*d.values())
        Printer.print(table)
    else:
        table = Printer.get_table(show_header=False, padding=0, box=box.SIMPLE)
        for k, v in dict_value.items():
            table.add_row(k, humanize_attrs(k, v))
        Printer.print(table)


class Printer:
    COLORS = ["yellow", "blue", "magenta", "green", "cyan", "red", "white"]
    console = Console(
        theme=Theme(
            {
                "header": "yellow",
                "success": "green",
                "info": "cyan",
                "warning": "magenta",
                "error": "red",
                "white": "white",
            }
        ),
        markup=True,
    )

    @staticmethod
    def get_progress():
        return Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            DownloadColumn(),
            TransferSpeedColumn(),
            TaskProgressColumn(),
            TextColumn("eta"),
            TimeRemainingColumn(),
            TextColumn("elapsed"),
            TimeElapsedColumn(),
        )

    @classmethod
    def get_live(cls):
        return Live(console=cls.console)

    @staticmethod
    def get_table(*args, **kwargs):
        return Table(*[Column(header=h, no_wrap=True) for h in args], **kwargs)

    @staticmethod
    def pprint(value):
        """Prints as formatted JSON"""
        click.echo(json.dumps(value, sort_keys=True, indent=4, separators=(",", ": ")))

    @classmethod
    def print_md(cls, md: str):
        cls.console.print(Markdown(md))

    @classmethod
    def print_text(cls, value: str):
        syntax = Syntax(value, "txt", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print_yaml(cls, value: any):
        if isinstance(value, str):
            value = yaml.safe_load(value)
        value = yaml.safe_dump(value, sort_keys=True, indent=2)
        syntax = Syntax(value, "yaml", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print_json(cls, value: any):
        if isinstance(value, str):
            value = json.loads(value)
        value = json.dumps(value, sort_keys=True, indent=4, separators=(",", ": "))
        syntax = Syntax(value, "json", theme="dracula", line_numbers=False)
        cls.console.print(syntax)

    @classmethod
    def print(cls, text):
        cls.console.print(text)

    @classmethod
    def help(cls, command_help: Optional[str] = None):
        if command_help:
            cls.console.print(
                "Please run [white]`polyaxon {} --help`[/white] for more details".format(
                    command_help
                ),
                style="info",
            )

    @classmethod
    def heading(cls, text):
        cls.header("\n{}\n".format(text))

    @classmethod
    def header(cls, text):
        cls.console.print(text, style="header")

    @classmethod
    def warning(cls, text, command_help: Optional[str] = None):
        cls.console.print(text, style="warning")
        if command_help:
            cls.help(command_help)

    @classmethod
    def success(cls, text):
        cls.console.print(text, style="success")

    @classmethod
    def error(
        cls, text, sys_exit: bool = False, command_help: Optional[str] = None, **kwargs
    ):
        cls.console.print(text, style="error")
        if command_help:
            cls.help(command_help)
        if sys_exit:
            sys.exit(1)

    @classmethod
    def tip(cls, text):
        cls.console.print(text, style="white")

    @classmethod
    def info(cls, text):
        cls.console.print(text, style="info")

    @staticmethod
    def add_log_color(value, color):
        return click.style("{}".format(value), fg=color)

    @classmethod
    def add_color(cls, value, style):
        return "[{style}]{value}[/{style}]".format(value=value, style=style)

    @classmethod
    def get_colored_status(cls, status):
        if status == "created":
            return cls.add_color(status, "info")
        elif status == "succeeded":
            return cls.add_color(status, style="success")
        elif status in ["failed", "stopped", "upstream_failed"]:
            return cls.add_color(status, style="error")
        elif status == "done":
            return cls.add_color(status, style="white")

        return cls.add_color(status, style="header")

    @classmethod
    def add_status_color(cls, obj_dict, status_key="status"):
        if obj_dict.get(status_key) is None:
            return obj_dict

        obj_dict[status_key] = cls.get_colored_status(obj_dict[status_key])
        return obj_dict

    @classmethod
    def add_memory_unit(cls, obj_dict, keys):
        keys = to_list(keys)
        for key in keys:
            obj_dict[key] = to_unit_memory(obj_dict[key])
        return obj_dict

    @classmethod
    def decorate_format_value(cls, text_format, values, color):
        values = to_list(values)
        values = [cls.add_color(value, color) for value in values]
        click.echo(text_format.format(*values))

    @staticmethod
    def log(value, nl=False):
        click.echo(value, nl=nl)

    @classmethod
    def resources(cls, jobs_resources):
        # TODO: move resources and other common configs to clippy
        from polyaxon.schemas.api.resources import ContainerResourcesConfig

        jobs_resources = to_list(jobs_resources)
        click.clear()
        table = Printer.get_table("Job", "Mem Usage / Total", "CPU% - CPUs")
        for job_resources in jobs_resources:
            job_resources = ContainerResourcesConfig.from_dict(job_resources)
            line = [
                job_resources.job_name,
                "{} / {}".format(
                    to_unit_memory(job_resources.memory_used),
                    to_unit_memory(job_resources.memory_limit),
                ),
                "{} - {}".format(
                    to_percentage(job_resources.cpu_percentage / 100),
                    job_resources.n_cpus,
                ),
            ]
            table.add_row(*line)
        Printer.print(table)
        sys.stdout.flush()

    @classmethod
    def gpu_resources(cls, jobs_resources):
        # TODO: move resources and other common configs to clippy
        from polyaxon.schemas.api.resources import ContainerResourcesConfig

        jobs_resources = to_list(jobs_resources)
        click.clear()
        table = Printer.get_table(
            "job_name",
            "name",
            "GPU Usage",
            "GPU Mem Usage / Total",
            "GPU Temperature",
            "Power Draw / Limit",
        )
        non_gpu_jobs = 0
        for job_resources in jobs_resources:
            job_resources = ContainerResourcesConfig.from_dict(job_resources)
            line = []
            if not job_resources.gpu_resources:
                non_gpu_jobs += 1
                continue
            for gpu_resources in job_resources.gpu_resources:
                line += [
                    job_resources.job_name,
                    gpu_resources.name,
                    to_percentage(gpu_resources.utilization_gpu / 100),
                    "{} / {}".format(
                        to_unit_memory(gpu_resources.memory_used),
                        to_unit_memory(gpu_resources.memory_total),
                    ),
                    gpu_resources.temperature_gpu,
                    "{} / {}".format(
                        gpu_resources.power_draw, gpu_resources.power_limit
                    ),
                ]
            table.add_row(*line)
        if non_gpu_jobs == len(jobs_resources):
            Printer.error(
                "No GPU job was found, please run `resources` command without `-g | --gpu` option."
            )
            exit(1)
        Printer.print(table)
        sys.stdout.flush()



from typing import List, Optional, Union

from clipped.list_utils import to_list


def validate_tags(
    tags: Optional[Union[str, List[str]]], validate_yaml: bool = False
) -> Optional[List[str]]:
    if not tags:
        return None

    if validate_yaml and isinstance(tags, str) and ("[" in tags and "]" in tags):
        import yaml

        tags = yaml.safe_load(tags)

    if isinstance(tags, str):
        tags = [tag.strip() for tag in tags.split(",")]
    tags = to_list(tags, to_unique=True)
    tags = [tag.strip() for tag in tags if (tag and isinstance(tag, str))]
    return [t for t in tags if t]



import os
import sys

from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand


def read_readme():
    if not os.path.exists("../README.md"):
        return ""
    with open("../README.md") as f:
        return f.read()


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errcode = pytest.main(self.test_args)
        sys.exit(errcode)


with open(os.path.join("./clipped/pkg.py"), encoding="utf8") as f:
    pkg = {}
    exec(f.read(), pkg)

setup(
    name=pkg["NAME"],
    version=pkg["VERSION"],
    description=pkg["DESC"],
    long_description=read_readme(),
    long_description_content_type="text/markdown",
    maintainer=pkg["AUTHOR"],
    maintainer_email=pkg["EMAIL"],
    author=pkg["AUTHOR"],
    author_email=pkg["EMAIL"],
    url=pkg["URL"],
    license=pkg["LICENSE"],
    platforms="any",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    keywords=[
        "cli",
        "configuration",
        "utils",
        "setup",
    ],
    python_requires=">=3.7",
    classifiers=[
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Typing :: Typed",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    tests_require=["pytest"],
    cmdclass={"test": PyTest},
)



import os
from typing import Dict, List, Optional

from clipped.cmd import run_command


def git_init(repo_path: str):
    run_command(
        cmd="git init {}".format(repo_path), data=None, location=repo_path, chw=True
    )


def update_submodules(repo_path: str):
    run_command(
        cmd="git submodule update --init --recursive --rebase --force",
        data=None,
        location=repo_path,
        chw=True,
    )


def git_fetch(repo_path: str, revision: str, flags: Optional[List[str]], env=None):
    flags = flags or []
    fetch_cmd = "git fetch {} origin".format(" ".join(flags))
    if revision:
        fetch_cmd = "{} {}".format(fetch_cmd, revision)
    fetch_cmd = "{} --update-head-ok --force".format(fetch_cmd)
    run_command(cmd=fetch_cmd, data=None, location=repo_path, chw=True, env=env)
    run_command(cmd="git checkout FETCH_HEAD", data=None, location=repo_path, chw=True)


def checkout_revision(repo_path: str, revision: str):
    """Checkout to a specific revision.

    If commit is None then checkout to master.
    """
    revision = revision or "master"
    run_command(
        cmd="git checkout {}".format(revision), data=None, location=repo_path, chw=True
    )


def add_remote(repo_path: str, url: str):
    run_command(
        cmd="git remote add origin {}".format(url),
        data=None,
        location=repo_path,
        chw=True,
    )


def set_remote(repo_path: str, url: str):
    run_command(
        cmd="git remote set-url origin {}".format(url),
        data=None,
        location=repo_path,
        chw=True,
    )


def get_status(repo_path: str):
    return run_command(cmd="git status -s", data=None, location=repo_path, chw=True)


def get_committed_files(repo_path: str, commit_hash: str):
    files_committed = run_command(
        cmd="git diff-tree --no-commit-id --name-only -r {}".format(commit_hash),
        data=None,
        location=repo_path,
        chw=True,
    ).split("\n")
    return [f for f in files_committed if f]


def git_undo(repo_path: str):
    run_command(cmd="git reset --hard", data=None, location=repo_path, chw=True)
    run_command(cmd="git clean -fd", data=None, location=repo_path, chw=True)


def git_commit(
    repo_path: str = ".",
    user_email: Optional[str] = None,
    user_name: Optional[str] = None,
    message: Optional[str] = None,
):
    message = message or "updated"
    run_command(cmd="git add -A", data=None, location=repo_path, chw=True)
    git_auth = "-c user.email=<{}> -c user.name={}".format(user_email, user_name)
    run_command(
        cmd='git {} commit -m "{}"'.format(git_auth, message),
        data=None,
        location=repo_path,
        chw=True,
    )


def is_git_initialized(path: str = ".") -> bool:
    return bool(
        run_command(
            cmd="git rev-parse --is-inside-work-tree",
            data=None,
            location=path,
            chw=True,
        ).split("\n")[0]
    )


def get_commit(path: str = ".") -> str:
    return run_command(
        cmd="git --no-pager log --pretty=oneline -1", data=None, location=path, chw=True
    ).split(" ")[0]


def get_head(path: str = ".") -> str:
    return run_command(
        cmd="git rev-parse HEAD", data=None, location=path, chw=True
    ).split("\n")[0]


def get_remote(repo_path: str = "."):
    current_remote = run_command(
        cmd="git config --get remote.origin.url",
        data=None,
        location=repo_path,
        chw=True,
    )
    return current_remote.strip("\n")


def get_repo_name(path: str = ".") -> str:
    repo = run_command(
        cmd="git rev-parse --show-toplevel", data=None, location=path, chw=True
    ).split("\n")[0]

    return os.path.basename(repo)


def get_branch_name(path: str = ".") -> str:
    return run_command(
        cmd="git rev-parse --abbrev-ref HEAD", data=None, location=path, chw=True
    ).split("\n")[0]


def is_dirty(path: str = ".") -> bool:
    return bool(run_command(cmd="git diff --stat", data=None, location=path, chw=True))


def get_code_reference(path: str = ".", url: Optional[str] = None) -> Dict:
    if not is_git_initialized(path):
        return {}

    url = url or get_remote(path)
    if "git@" in url:
        url = url.split("git@")[1]
        url = url.split(".git")[0]
        url = url.replace(":", "/")
        url = "https://" + url
    return {"commit": get_commit(path), "branch": get_branch_name(path), "url": url}


def get_code_reference_all(path: str = ".") -> Dict:
    if not is_git_initialized(path):
        return {}

    return {
        "commit": get_commit(path),
        "head": get_head(path),
        "branch": get_branch_name(path),
        "url": get_remote(path),
        "is_dirty": is_dirty(path),
    }
